

#include "softshell.cu"
#include "tools/utils.h"
#include "timing.h"
#include "clustering.h"
#include "helpers.cuh"
#include "scheduled_updateNode.cuh"

namespace Softshell
{
  Timing timing;

  Scheduler scheduling;
  
  template<uint ERROR_DEPTH>
  size_t buildMesh(CUDAStorages* memory, float threshold, Event<TraverseEvent<ERROR_DEPTH> >* useevent);

  Event<TraverseEvent<0> > *buildEvent0;
  Event<TraverseEvent<1> > *buildEvent1;
  
  unsigned int usedDevice;

  void initScheduling(int device)
  {
    usedDevice = device;
    scheduler.init<TraversePriority>(usedDevice);

    buildEvent0 = scheduler.createEvent<TraverseEvent<0> >();
    buildEvent1 = scheduler.createEvent<TraverseEvent<1> >();
  }
  void shutdownScheduling()
  {
    scheduler.shutdown();
  }

  size_t buildMesh(CUDAStorages* memory, float threshold, bool error_depth)
  {
    if(error_depth)
      return buildMesh<1>(memory, threshold, buildEvent0);
    else
      return buildMesh<0>(memory, threshold, buildEvent1);
  }

  void prepareData(CUDAStorages* memory)
  {
     setOctreeBoundaries(memory->boundaries);
     setLevelOffsets(memory->h_levelOffset.get());

     CUDA_CHECKED_CALL(cudaMemcpyToSymbol(octree, &memory->d_octree_data, sizeof(OctreeEntry*)));
     CUDA_CHECKED_CALL(cudaMemcpyToSymbol(vertexCoherence, &memory->d_octree_vertex_coherence, sizeof(const unsigned int*)));
     CUDA_CHECKED_CALL(cudaMemcpyToSymbol(vertexIdOut, &memory->d_vertex_lookup, sizeof(unsigned int *)));
     CUDA_CHECKED_CALL(cudaMemcpyToSymbol(vertices, &memory->vertexcount, sizeof(unsigned int)));
     CUDA_CHECKED_CALL(cudaMemcpyToSymbol(maxdepth, &memory->maxdepth, sizeof(unsigned int)));
     CUDA_CHECKED_CALL(cudaMemcpyToSymbol(voxelIds, &memory->d_voxel_ids, sizeof(const uint3*)));
     CUDA_CHECKED_CALL(cudaMemcpyToSymbol(triangleCoherence, &memory->d_octree_triangle_coherence, sizeof(const unsigned int*)));
  }

  template<uint ERROR_DEPTH>
  size_t buildMesh(CUDAStorages* memory, float threshold, Event<TraverseEvent<ERROR_DEPTH> >* useevent)
  {
    timing.begin();

    uint null = 0;
    CUDA_CHECKED_CALL(cudaMemcpyToSymbol(triangleOutCount, &null, sizeof(uint)));
    CUDA::GL::mapped_buffer<unsigned int> d_indices(memory->cuda_index_buffer);
    CUDA_CHECKED_CALL(cudaMemcpyToSymbol(triOutput, &d_indices.get(), sizeof(uint*)));

    useevent->fire(threshold);
    useevent->join();

    uint triangles;
    CUDA_CHECKED_CALL(cudaMemcpyFromSymbol(&triangles, triangleOutCount, sizeof(uint)));

    double dt = timing.end();
    std::cout << "Softshell: " << dt*1000 << " (" << triangles << " triangles)" <<  std::endl;

    return 3*triangles;
  }
};
